#!/usr/bin/env python

"""Tests for `md_planning` package."""

import os
import pathlib
import pytest
import datetime
import re
import pandas as pd
import yaml

from click.testing import CliRunner

from md_planning.md_planning import (
    GanttDrawer,
    PertDrawer,
    ProjectReader,
    rename,
    get_task_type,
    is_nested,
    walk_project_tasks,
)

from gantt import Resource, Task

from md_planning import cli


# TODO: (feature) critical_color implementation
# TODO: (bugfix) E17FB6 v0.2 error message shows what csv task and what is the offensive entry
# BUG: if there is not at least one dependency between tasks then python-gantt does NOT create the svg
# TODO: (feature) assert that task names in a multiproject definition are unique for the whole definition file
# TODO: (refactor) data structure and project classes so the Resource class has methods: get_price, get_unit
# TODO: (refactor) data structure and project classes so the Task class has method: get_usage
# TODO: (refactor) data structure and project classes so the Project class has methods: get_unit, get_price, get_usage, get_cost
# TODO: (feature) when user makes task depend on milestone, KeyError is raised but it is difficult to find in which task the error was made
# TODO: (bug) nested definition implementation supposes that the CP of a project is the `SUM` of CPs of all it's milestones. This is not true if a "small" task in a milestone depends on a long running task outside of the current milestone. The only way to solve this now is to build the full CP between milestones.
# TODO: (refactor) add get_dependencies and get_resources to the main code

@pytest.fixture
def flat_project():
    project_str = """
# Font:
#     fill: "#000000"
#     stroke: "black"
#     stroke_width: 0
#     font_family: Verdana
Vacations:
    - 2022-09-30
    - 2022-11-01
    - 2022-11-11
Resources:
    Martin:
        price:
            value: 68.75
            unit: workhour
        vacations:
            - 2022-09-15
    Samuel:
        price:
            value: 600
            unit:  workday
Projects:
    -   Name: Test1
        Tasks:
            kickoff:
                type: milestone
                depends_on: brief
            brief:
                type: task
                start: 2022-09-05
                duration: 0.125 # 1 hour
                percent_done: 0
                resources: null
                depends_on: null
            goals:
                type: task
                start: 2022-09-05
                duration: 0.25
                percent_done: 0
                resources: Martin
                depends_on: brief
            Env setup: [2022-09-06, 1, 0, "Martin", "goals"]
"""
    return project_str


@pytest.fixture
def nested_project():
    project_str = """
# Font:
#     fill: black
#     font_family: Verdana
Vacations:
    - 2022-09-30
    - 2022-11-01
    - 2022-11-11
Resources:
    Martin:
        price:
            value: 68.75
            unit: workhour
        vacations:
            - 2022-09-15
    Samuel:
        price:
            value: 600
            unit:  workday
Projects:
    -   Name: Test1
        Tasks:
            kickoff:
                brief:
                    type: task
                    start: 2022-09-05
                    duration: 0.125 # 1 hour
                    percent_done: 0
                    resources: null
                    depends_on: null
            goals:
                type: task
                start: 2022-09-05
                duration: 0.25
                percent_done: 0
                resources: Martin
                depends_on: brief
            Env setup: [2022-09-06, 1, 0, "Martin", "goals"]
"""
    return project_str


@pytest.fixture
def nested_error_project():
    project_str = """
# Font:
#     fill: black
#     font_family: Verdana
Vacations:
    - 2022-09-30
    - 2022-11-01
    - 2022-11-11
Resources:
    Martin:
        price:
            value: 68.75
            unit: workhour
        vacations:
            - 2022-09-15
    Samuel:
        price:
            value: 600
            unit:  workday
Projects:
    -   Name: Test1
        Tasks:
            kickoff:
                brief:
                    type: task
                    start: 2022-09-05
                    duration: null
                    percent_done: 0
                    resources: null
                    depends_on: null
                    best: null
                    optimal: 5
                    worst: 10
            goals:
                type: task
                start: 2022-09-05
                duration: 0.25
                percent_done: 0
                resources: Martin
                depends_on: brief
            Env setup: [2022-09-06, 1, 0, "Martin", "goals"]
"""
    return project_str


################################################################################
# UTILITY FUNCTIONS
################################################################################


def test_rename_strict():
    with pytest.raises(KeyError):
        _ = dict()
        rename(_, "a", "b")


def test_rename_flex_exists():
    _ = dict(a=1)
    res = rename(_, "a", "b", True)
    assert res == {"b": 1}


def test_rename_flex_not_exists():
    _ = dict()
    res = rename(_, "a", "b", True, 1)
    assert res == {"b": 1}


def test_get_task_type_flat_milestone():
    milestone = dict(type="milestone", depends_on=[])
    assert get_task_type(milestone) == "milestone"


def test_get_task_type_flat_task_dict():
    task = {
        "type": "task",
        "start": "",
        "duration": 1,
        "percent_done": 0,
        "resources": [],
        "depends_on": [],
        "color": None,
    }
    assert get_task_type(task) == "task"


def test_get_task_type_flat_task_list():
    task = ["2022-09-06", 1, 0, "Martin", "kickoff", None]
    assert get_task_type(task) == "task"


def test_get_task_type_nested_milestone():
    milestone = dict(type="milestone", depends_on=[])
    assert get_task_type(milestone) == "milestone"


def test_get_task_type_nested_task_dict():
    task = {
        "start": "",
        "duration": 1,
        "percent_done": 0,
        "resources": [],
        "depends_on": [],
        "color": None,
    }
    assert get_task_type(task) == "task"


def test_get_task_type_nested_task_list():
    task = ["2022-09-06", 1, 0, "Martin", "kickoff", None]
    assert get_task_type(task) == "task"


def test_is_nested_flat_is_false(flat_project):
    proj = yaml.full_load(flat_project)
    assert is_nested(proj["Projects"][0]["Tasks"]) == False


def test_is_nested_nested_is_true(nested_project):
    proj = yaml.full_load(nested_project)
    assert is_nested(proj["Projects"][0]["Tasks"]) == True


def test_walk_project_returns_all_tasks(nested_project):
    proj = yaml.full_load(nested_project)
    walked = walk_project_tasks(proj["Projects"][0]["Tasks"])
    assert isinstance(walked, list)
    assert (
        all([key in ["kickoff", "brief", "goals", "Env setup"] for (key, _) in walked])
        and len(walked) > 0
    )


################################################################################
# PROJECTREADER
################################################################################


def test_project_reader_build_flat_project_is_conformant(flat_project):
    p = ProjectReader(flat_project)
    projects, pertcharts = p.build_project()

    assert len(pertcharts) == 1
    assert len(projects["projects"]) == 1
    assert pertcharts == {
        "Test1": {
            "Env setup": {
                "Tid": "Env setup",
                "start": 0,
                "duration": 1,
                "end": 0,
                "responsible": "CRITICAL",
                "pred": ["goals"],
            },
            "brief": {
                "Tid": "brief",
                "start": 0,
                "duration": 0.125,
                "end": 0,
                "responsible": "CRITICAL",
                "pred": ["START"],
            },
            "goals": {
                "Tid": "goals",
                "start": 0,
                "duration": 0.25,
                "end": 0,
                "responsible": "CRITICAL",
                "pred": ["brief"],
            },
            "kickoff": {
                "Tid": "kickoff",
                "start": 0,
                "duration": 0,
                "end": 0,
                "pred": ["brief"],
                "responsible": "",

            },
        }
    }
    assert projects == {
        # "font": {"fill": "black", "font_family": "Verdana"},
        "font": {},
        "vacations": [
            datetime.date(2022, 9, 30),
            datetime.date(2022, 11, 1),
            datetime.date(2022, 11, 11),
        ],
        "resources": {
            "Martin": {
                "price": {"value": 68.75, "unit": "workhour"},
                "vacations": [datetime.date(2022, 9, 15)],
            },
            "Samuel": {"price": {"value": 600, "unit": "workday"}},
        },
        "projects": {"Test1": None},
        "tasks": {
            "kickoff": {
                "project": "Test1",
                "task": {
                    "name": "kickoff",
                    "start": None,
                    "stop": None,
                    "duration": 0,
                    "depends_of": ["brief"],
                    "resources": None,
                    "percent_done": 0,
                    "color": None,
                    "fullname": None,
                    "display": True,
                    "state": "",
                },
                "milestone": {
                    "name": "kickoff",
                    "start": None,
                    "depends_of": ["brief"],
                    "color": None,
                    "fullname": None,
                    "display": True,
                },
                "pertchart": {
                    "Tid": "kickoff",
                    "start": 0,
                    "duration": 0,
                    "end": 0,
                    "pred": ["brief"],
                    "responsible": "",
                },
                "cpm": {
                    "name": "kickoff",
                    "duration": 0,
                    "lag": 0,
                    "depends_of": ["brief"],
                },
                "is_milestone": True,
            },
            "brief": {
                "project": "Test1",
                "task": {
                    "name": "brief",
                    "start": datetime.date(2022, 9, 5),
                    "stop": None,
                    "duration": 0.125,
                    "depends_of": [],
                    "resources": [],
                    "percent_done": 0,
                    "color": None,
                    "fullname": None,
                    "display": True,
                    "state": "",
                },
                "milestone": {
                    "name": "brief",
                    "start": None,
                    "depends_of": None,
                    "color": None,
                    "fullname": None,
                    "display": True,
                },
                "pertchart": {
                    "Tid": "brief",
                    "start": 0,
                    "duration": 0.125,
                    "end": 0,
                    "responsible": "CRITICAL",
                    "pred": ["START"],
                },
                "cpm": {"name": "brief", "duration": 0.125, "lag": 0, "depends_of": []},
                "is_milestone": False,
            },
            "goals": {
                "project": "Test1",
                "task": {
                    "name": "goals",
                    "start": datetime.date(2022, 9, 5),
                    "stop": None,
                    "duration": 0.25,
                    "depends_of": ["brief"],
                    "resources": ["1 Martin"],
                    "percent_done": 0,
                    "color": None,
                    "fullname": None,
                    "display": True,
                    "state": "",
                },
                "milestone": {
                    "name": "goals",
                    "start": None,
                    "depends_of": None,
                    "color": None,
                    "fullname": None,
                    "display": True,
                },
                "pertchart": {
                    "Tid": "goals",
                    "start": 0,
                    "duration": 0.25,
                    "end": 0,
                    "responsible": "CRITICAL",
                    "pred": ["brief"],
                },
                "cpm": {
                    "name": "goals",
                    "duration": 0.25,
                    "lag": 0,
                    "depends_of": ["brief"],
                },
                "is_milestone": False,
            },
            "Env setup": {
                "project": "Test1",
                "task": {
                    "name": "Env setup",
                    "start": datetime.date(2022, 9, 6),
                    "stop": None,
                    "duration": 1,
                    "depends_of": ["goals"],
                    "resources": ["1 Martin"],
                    "percent_done": 0,
                    "color": None,
                    "fullname": None,
                    "display": True,
                    "state": "",
                },
                "milestone": {
                    "name": "Env setup",
                    "start": None,
                    "depends_of": None,
                    "color": None,
                    "fullname": None,
                    "display": True,
                },
                "pertchart": {
                    "Tid": "Env setup",
                    "start": 0,
                    "duration": 1,
                    "end": 0,
                    "responsible": "CRITICAL",
                    "pred": ["goals"],
                },
                "cpm": {
                    "name": "Env setup",
                    "duration": 1,
                    "lag": 0,
                    "depends_of": ["goals"],
                },
                "is_milestone": False,
            },
        },
    }


def test_project_reader_build_nested_project_is_conformant(nested_project):
    p = ProjectReader(nested_project)
    projects, pertcharts = p.build_project()

    assert len(pertcharts) == 1
    assert len(projects["projects"]) == 1
    assert pertcharts == {
        "Test1": {
            "Env setup": {
                "Tid": "Env setup",
                "start": 0,
                "duration": 1,
                "end": 0,
                "responsible": "CRITICAL",
                "pred": ["goals"],
            },
            "brief": {
                "Tid": "brief",
                "start": 0,
                "duration": 0.125,
                "end": 0,
                "responsible": "CRITICAL",
                "pred": ["START"],
            },
            "goals": {
                "Tid": "goals",
                "start": 0,
                "duration": 0.25,
                "end": 0,
                "responsible": "CRITICAL",
                "pred": ["brief"],
            },
            "kickoff": {
                "Tid": "kickoff",
                "start": 0,
                "duration": 0,
                "end": 0,
                "pred": ["brief"],
                "responsible": "",
            },
        }
    }
    assert projects == {
        # "font": {"fill": "black", "font_family": "Verdana"},
        "font": {},
        "vacations": [
            datetime.date(2022, 9, 30),
            datetime.date(2022, 11, 1),
            datetime.date(2022, 11, 11),
        ],
        "resources": {
            "Martin": {
                "price": {"value": 68.75, "unit": "workhour"},
                "vacations": [datetime.date(2022, 9, 15)],
            },
            "Samuel": {"price": {"value": 600, "unit": "workday"}},
        },
        "projects": {"Test1": None},
        "tasks": {
            "kickoff": {
                "project": "Test1",
                "task": {
                    "name": "kickoff",
                    "start": None,
                    "stop": None,
                    "duration": 0,
                    "depends_of": ["brief"],
                    "resources": None,
                    "percent_done": 0,
                    "color": None,
                    "fullname": None,
                    "display": True,
                    "state": "",
                },
                "milestone": {
                    "name": "kickoff",
                    "start": None,
                    "depends_of": ["brief"],
                    "color": None,
                    "fullname": None,
                    "display": True,
                },
                "pertchart": {
                    "Tid": "kickoff",
                    "start": 0,
                    "duration": 0,
                    "end": 0,
                    "pred": ["brief"],
                    "responsible": "",
                },
                "cpm": {
                    "name": "kickoff",
                    "duration": 0,
                    "lag": 0,
                    "depends_of": ["brief"],
                },
                "is_milestone": True,
            },
            "brief": {
                "project": "Test1",
                "task": {
                    "name": "brief",
                    "start": datetime.date(2022, 9, 5),
                    "stop": None,
                    "duration": 0.125,
                    "depends_of": [],
                    "resources": [],
                    "percent_done": 0,
                    "color": None,
                    "fullname": None,
                    "display": True,
                    "state": "",
                },
                "milestone": {
                    "name": "brief",
                    "start": None,
                    "depends_of": None,
                    "color": None,
                    "fullname": None,
                    "display": True,
                },
                "pertchart": {
                    "Tid": "brief",
                    "start": 0,
                    "duration": 0.125,
                    "end": 0,
                    "responsible": "CRITICAL",
                    "pred": ["START"],
                },
                "cpm": {"name": "brief", "duration": 0.125, "lag": 0, "depends_of": []},
                "is_milestone": False,
            },
            "goals": {
                "project": "Test1",
                "task": {
                    "name": "goals",
                    "start": datetime.date(2022, 9, 5),
                    "stop": None,
                    "duration": 0.25,
                    "depends_of": ["brief"],
                    "resources": ["1 Martin"],
                    "percent_done": 0,
                    "color": None,
                    "fullname": None,
                    "display": True,
                    "state": "",
                },
                "milestone": {
                    "name": "goals",
                    "start": None,
                    "depends_of": None,
                    "color": None,
                    "fullname": None,
                    "display": True,
                },
                "pertchart": {
                    "Tid": "goals",
                    "start": 0,
                    "duration": 0.25,
                    "end": 0,
                    "responsible": "CRITICAL",
                    "pred": ["brief"],
                },
                "cpm": {
                    "name": "goals",
                    "duration": 0.25,
                    "lag": 0,
                    "depends_of": ["brief"],
                },
                "is_milestone": False,
            },
            "Env setup": {
                "project": "Test1",
                "task": {
                    "name": "Env setup",
                    "start": datetime.date(2022, 9, 6),
                    "stop": None,
                    "duration": 1,
                    "depends_of": ["goals"],
                    "resources": ["1 Martin"],
                    "percent_done": 0,
                    "color": None,
                    "fullname": None,
                    "display": True,
                    "state": "",
                },
                "milestone": {
                    "name": "Env setup",
                    "start": None,
                    "depends_of": None,
                    "color": None,
                    "fullname": None,
                    "display": True,
                },
                "pertchart": {
                    "Tid": "Env setup",
                    "start": 0,
                    "duration": 1,
                    "end": 0,
                    "responsible": "CRITICAL",
                    "pred": ["goals"],
                },
                "cpm": {
                    "name": "Env setup",
                    "duration": 1,
                    "lag": 0,
                    "depends_of": ["goals"],
                },
                "is_milestone": False,
            },
        },
    }


def test_project_reader_nested_raises_exceptions_with_task_name(nested_error_project):
    with pytest.raises(ValueError) as err:
        p = ProjectReader(nested_error_project)
    assert "brief" in str(err.from_exc_info(err._excinfo))


################################################################################
# PERTDRAWER
################################################################################


def test_pertdrawer_draw_creates_chart_file(flat_project):
    project_def = ProjectReader(flat_project)
    _, pert_def = project_def.build_project()
    pert = PertDrawer(pert_def)
    pert.draw()
    assert pathlib.Path("Test1_pert.pdf").is_file()
    os.unlink("Test1_pert.pdf")
    os.unlink("Test1_pert.gv")


################################################################################
# GANTTDRAWER
################################################################################


def test_gantt_drawer_creates_gantt_svg_output(flat_project):
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)
    gantt.draw_tasks(project="Test1")
    assert pathlib.Path("Test1.svg").is_file()
    os.unlink("Test1.svg")


def test_gantt_drawer_creates_resource_svg_output(flat_project):
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)
    gantt.draw_resources(project="Test1")
    assert pathlib.Path("Test1_resources.svg").is_file()
    os.unlink("Test1_resources.svg")


################################################################################
# Resource budgeting feature
################################################################################


def test_resource_has_attribute_price(flat_project):
    # refactoring of the resource schema to change Resource.cost to Resource.price
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)

    assert hasattr(gantt.resources["Martin"], "price")
    assert hasattr(gantt.resources["Samuel"], "price")


def test_gantt_reader_can_define_new_units(flat_project):
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)
    gantt.define_unit("foo = 1 * hour = ⧘ = wiggly_foo")
    assert "foo" in gantt.registry
    assert "wiggly_foo" in gantt.registry


def test_gantt_reader_can_make_units_alias(flat_project):
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)

    # define abbreviated syntax
    gantt.define_alias("workday = foo2")
    assert "foo2" in gantt.registry
    assert gantt.registry.foo2 == gantt.registry.workday

    # define full syntax
    gantt.define_alias("@alias workday = foo3")
    assert "foo3" in gantt.registry
    assert gantt.registry.foo3 == gantt.registry.workday


def test_project_reader_reads_strips_spaces_from_task_resources():
    # make sure that all resource quantifiers are present as necessary
    project_str = """
Vacations:
    - 2022-09-30
    - 2022-11-01
    - 2022-11-11
Resources:
    Martin:
        price:
            value: 68.75
            unit: hour
        vacations:
            - 2022-09-15
    Samuel:
        price:
            value: 600
            unit:  day
Projects:
    -   Name: Test1
        Tasks:
            kickoff:
                type: milestone
                depends_on: brief
            brief:
                type: task
                start: 2022-09-05
                duration: 0.125 # 1 hour
                percent_done: 0
                resources: 0.5 Martin, 2 Samuel
                depends_on: null
            goals:
                type: task
                start: 2022-09-05
                duration: 0.25
                percent_done: 0
                resources: "Martin "
                depends_on: brief
            Env setup: [2022-09-06, 1, 0, "Martin", "goals"]
"""
    project_def = ProjectReader(project_str)
    project_def, _ = project_def.build_project()

    brief = project_def["tasks"]["brief"]["task"]
    assert brief["resources"] == ["0.5 Martin", "2 Samuel"]

    goals = project_def["tasks"]["goals"]["task"]
    assert goals["resources"] == ["1 Martin"]


def test_gantt_reader_supports_scaled_resources():
    # DONE: gantt.tasks[name].resources must contain a gantt.Resource instance
    # DONE: gantt.data.project[name].tasks[name].resources must contain the scaled string
    project_str = """
Vacations:
    - 2022-09-30
    - 2022-11-01
    - 2022-11-11
Resources:
    Martin:
        price:
            value: 68.75
            unit: hour
        vacations:
            - 2022-09-15
    Samuel:
        price:
            value: 600
            unit:  day
Projects:
    -   Name: Test1
        Tasks:
            kickoff:
                type: milestone
                depends_on: brief
            brief:
                type: task
                start: 2022-09-05
                duration: 0.125 # 1 hour
                percent_done: 0
                resources: 0.5 Martin, 2 Samuel
                depends_on: null
            goals:
                type: task
                start: 2022-09-05
                duration: 0.25
                percent_done: 0
                resources: Martin
                depends_on: brief
            Env setup: [2022-09-06, 1, 0, "Martin", "goals"]
"""
    project = ProjectReader(project_str)
    gantt_def, _ = project.build_project()

    gantt = GanttDrawer(gantt_def)

    # scaled string: digit + resource identifier, eg" "1 Martin", "0.5 Samuel"
    # regex = r"^(\d+\.?\d*)\s+(\w+)$"
    try:
        regex = re.compile(r"^(\d+\.?\d*)\s+(\w+)$")
        assert [
            all(
                [
                    float(regex.match(r).group(1)) and bool(regex.match(r).group(2))
                    for r in t_value["task"]["resources"]
                    if regex.match(r)
                ]
            )
            for t_name, t_value in gantt.data["tasks"].items()
            if not gantt.data["tasks"][t_name]["is_milestone"]
        ]
    except ValueError as exc:
        assert False, f"Could not convert quantifier to number {exc}"
    except KeyError as exc:
        assert False, f"Could not find resource: {exc}"

    assert (
        all(
            [
                all([isinstance(r, Resource) for r in t_value.resources])
                for t_name, t_value in gantt.tasks.items()
                if t_value.__class__.__name__ == "Task"
            ]
        )
        and len(gantt.tasks) > 0
    )  # NOTE: milestone instance isinstance (..., Task) == True so must discriminate by class name


def test_resource_instance_is_monkey_patched_get_unit(flat_project):
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)
    assert hasattr(gantt.resources["Martin"], "get_unit")
    assert hasattr(gantt.resources["Samuel"], "get_unit")
    assert gantt.resources["Martin"].get_unit() == "workhour"
    assert gantt.resources["Samuel"].get_unit() == "workday"


@pytest.mark.skip("later version")
def test_resource_instance_has_category_attribute(flat_project):
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)
    assert hasattr(gantt.resources["Martin"], "category")
    assert hasattr(gantt.resources["Samuel"], "category")
    assert gantt.resources["Martin"].category is None
    assert gantt.resources["Samuel"].category is None


@pytest.mark.skip("later version")
def test_resource_instance_has_subcategory_attribute(flat_project):
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)
    assert hasattr(gantt.resources["Martin"], "subcategory")
    assert hasattr(gantt.resources["Samuel"], "subcategory")
    assert gantt.resources["Martin"].subcategory is None
    assert gantt.resources["Samuel"].subcategory is None


def test_resource_instance_is_monkey_patched_get_price(flat_project):
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)
    assert hasattr(gantt.resources["Martin"], "get_price")
    assert hasattr(gantt.resources["Samuel"], "get_price")
    assert gantt.resources["Martin"].get_price() == 68.75
    assert gantt.resources["Samuel"].get_price() == 600


def test_resource_instance_is_monkey_patched_convert(flat_project):
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)
    assert hasattr(gantt.resources["Martin"], "convert")
    assert hasattr(gantt.resources["Samuel"], "convert")
    assert gantt.resources["Martin"].convert("workday") == 550
    assert gantt.resources["Samuel"].convert("workday") == 600
    assert gantt.resources["Samuel"].convert("workweek") == pytest.approx(3000)


def test_resource_instance_is_monkey_patched_forbid_holidays_for_some_units():
    # if a resource has unit unit/batch then raise valueerror if it defines a holiday
    # if a batch resource does not define a batch_size raise valueerror
    proj1_str = project_str = """
Vacations:
    - 2022-09-30
    - 2022-11-01
    - 2022-11-11
Resources:
    Martin:
        price:
            value: 68.75
            unit: unit
        vacations:
            - 2022-09-15
    Samuel:
        price:
            value: 600
            unit:  workday
Projects:
    -   Name: Test1
        Tasks:
            kickoff:
                type: milestone
                depends_on: brief
            brief:
                type: task
                start: 2022-09-05
                duration: 0.125 # 1 hour
                percent_done: 0
                resources: null
                depends_on: null
            goals:
                type: task
                start: 2022-09-05
                duration: 0.25
                percent_done: 0
                resources: Martin
                depends_on: brief
            Env setup: [2022-09-06, 1, 0, "Martin", "goals"]
"""
    proj1_def = ProjectReader(proj1_str).build_project()[0]
    with pytest.raises(ValueError):
        gantt = GanttDrawer(proj1_def)

    proj2_str = project_str = """
# Font:
#     fill: "#000000"
#     stroke: "black"
#     stroke_width: 0
#     font_family: Verdana
Vacations:
    - 2022-09-30
    - 2022-11-01
    - 2022-11-11
Resources:
    Martin:
        price:
            value: 68.75
            unit: batch
    Samuel:
        price:
            value: 600
            unit:  workday
Projects:
    -   Name: Test1
        Tasks:
            kickoff:
                type: milestone
                depends_on: brief
            brief:
                type: task
                start: 2022-09-05
                duration: 0.125 # 1 hour
                percent_done: 0
                resources: null
                depends_on: null
            goals:
                type: task
                start: 2022-09-05
                duration: 0.25
                percent_done: 0
                resources: Martin
                depends_on: brief
            Env setup: [2022-09-06, 1, 0, "Martin", "goals"]
"""
    proj2_def = ProjectReader(proj2_str).build_project()[0]
    with pytest.raises(ValueError):
        gantt = GanttDrawer(proj2_def)


def test_task_instance_is_monkeypatched_is_using(flat_project):
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)
    assert hasattr(gantt.tasks["goals"], "is_using")
    assert gantt.tasks["goals"].is_using("Martin") == True
    assert gantt.tasks["goals"].is_using("Samuel") == False


def test_task_instance_is_monkeypatched_is_active(flat_project):
    # a task is active on the day when it is running.
    # fractional duration counts for a full day in accounting for activity.
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)
    assert hasattr(gantt.tasks["goals"], "is_active")
    # Notice the date is 2022-09-06 and not 2022-09-05 as defined in the project
    # this is because "goals" task is dependent on "kickoff" task and
    # gantt automatically plans for the depending task 1 day after parent task
    # gantt is not schedule optimization software and the resolution is 1 day
    # we keep these design choices for now.
    # notice the dependency has precedence over the the start/stop date argument
    assert gantt.tasks["goals"].is_active("2022-09-06") == True
    assert gantt.tasks["goals"].is_active("2022-09-07") == False
    assert hasattr(gantt.tasks["Env setup"], "is_active")
    assert gantt.tasks["Env setup"].is_active("2022-09-07") == True
    assert gantt.tasks["Env setup"].is_active("2022-09-08") == False


def test_ganttdrawer_get_unit(flat_project):
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)
    assert hasattr(gantt, "get_unit")
    assert gantt.get_unit("Martin") == "workhour"
    assert gantt.get_unit("Samuel") == "workday"


def test_ganttdrawer_get_price(flat_project):
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)
    assert hasattr(gantt, "get_price")
    assert gantt.get_price("Martin") == 68.75
    assert gantt.get_price("Samuel") == 600


def test_ganttdrawer_get_usage(flat_project):
    # return the quantity that a resource is used in a specified task in a project
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)
    assert hasattr(gantt, "get_usage")
    assert gantt.get_usage("goals", "Martin") == 1
    assert gantt.get_usage("goals", "Samuel") == 0


def test_ganttdrawer_is_available(flat_project):
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)
    assert hasattr(gantt, "is_available")

    # wrong date
    with pytest.raises(ValueError):
        gantt.is_available("Martin", "")
    # unavailable date within project range
    assert gantt.is_available("Martin", "2022-09-08") == False
    # available date
    assert gantt.is_available("Martin", "2022-09-05") == True


def test_ganttdrawer_is_using(flat_project):
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)
    assert hasattr(gantt, "is_using")

    # unavailable date within project range
    assert gantt.is_using("goals", "Samuel") == False
    # available date
    assert gantt.is_using("goals", "Martin") == True


def test_ganttdrawer_get_cost(flat_project):
    # gantt.convert(resource) * gantt.get_usage(task, resource) * gantt.is_available(resource, date) * gantt.is_using(resource) * gantt.is_active(task, date)
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)
    assert hasattr(gantt, "get_cost")

    # unavailable date within project range
    assert gantt.get_cost("goals", "Martin", "2022-09-06") == 137.5
    assert gantt.get_cost("goals", "Samuel", "2022-09-06") == 0


def test_ganttdrawer_budget_flat_def(flat_project):
    project_def = ProjectReader(flat_project)
    project_def, _ = project_def.build_project()
    gantt = GanttDrawer(project_def)
    assert hasattr(gantt, "budget")

    # unavailable date within project range

    check = pd.DataFrame(
        [
            ["Test1", "goals", "Martin", "", "", datetime.date(2022, 9, 6), 137.5],
            ["Test1", "Env setup", "Martin", "", "", datetime.date(2022, 9, 7), 550.0],
        ],
        columns=[
            "project",
            "task",
            "resource",
            "category",
            "subcategory",
            "date",
            "amount",
        ],
    )

    assert gantt.budget().equals(check)


################################################################################
# \BUGFIXES
################################################################################


def test_patch_bug_C50565():
    """:BUGFIX: C50565 a project without any task dependency does not produce project outputs"""
    project_str = """
Vacations:
    - 2022-09-30
Resources:
    Martin:
        price:
            value: 68.75
            unit: hour
        vacations:
            - 2022-09-15
    Julie:
        price:
            value: 80
            unit: hour
        vacations:
            - 2022-09-15
Projects:
    -   Name: test_project_1
        Tasks:
            brief:
                type: task
                start: 2022-09-05
                duration: 0.125
                percent_done: 0
                resources: null
                depends_on: null
            test:
                type: task
                start: 2022-09-06
                duration: 0.25
                percent_done: 0
                resources: Martin,Julie
                depends_on: brief
"""

    gantt_def, _ = ProjectReader(project_str).build_project()

    # pert = PertDrawer(pert_def)
    # pert.draw()

    gantt = GanttDrawer(gantt_def)
    gantt.draw_tasks(project="test_project_1")
    assert pathlib.Path("test_project_1.svg").is_file()
    os.unlink("test_project_1.svg")


def test_patch_bug_E17FB6():
    """:BUGFIX: E17FB6 wrong csv task definition prevents project creation"""
    project_str = """
Vacations: #general
    - 2022-09-30
Resources:
    Martin:
        price:
            value: 68.75
            unit: hour
        vacations:
            - 2022-09-15
    Julie:
        price:
            value: 80
            unit: hour
        vacations:
            - 2022-09-15
Projects:
    -   Name: test_project_2
        Tasks:
            kickoff:
                type: milestone
                depends_on: brief
            brief:
                type: task
                start: 2022-09-05
                duration: 0.125
                percent_done: 0
                resources: []
                depends_on: []
            # order: ["start", "duration", "percent_done", "resources", "depends_on", "color"]
            Env setup: [2022-09-06, 1, 0, "Martin", "brief"]
"""

    gantt_def, _ = ProjectReader(project_str).build_project()

    # pert = PertDrawer(pert_def)
    # pert.draw()

    gantt = GanttDrawer(gantt_def)
    gantt.draw_tasks(project="test_project_2")
    assert pathlib.Path("test_project_2.svg").is_file()
    os.unlink("test_project_2.svg")


def test_patch_bug_A14E36():
    """
    :BUGFIX: A14E36 pertdrawer returns error when renaming the pert drawings

    sys.stdout shows message: "The file /Users/martinteller/Documents/l1nxit/Documentation/PERT.gv.pdf does not exist."

    - unable to recreate the bug from source code
    - insert breakpoint() inside the source code to attach to the running process launched from markdown-preview-enhanced: unable to reproduce the bug from running subprocess
    - try condisitonal move pathlib.Path(...).is_file() --(true)--> shutil.move(...) UNSUCCESSFUL

    status: PENDING
    """
    project_str = """
Vacations: #general
    - 2022-09-30
Resources:
    Martin:
        price:
            value: 68.75
            unit: hour
        vacations: #for a resource
            - 2022-09-15
    Julie:
        price:
            value: 80
            unit: hour
        vacations: #for a resource
            - 2022-09-15
Projects:
    -   Name: test_project_3
        Tasks:
            kickoff:
                type: milestone
                depends_on: brief
            brief:
                type: task
                start: 2022-09-05
                duration: 0.125 # 1 hour
                percent_done: 0
                resources: null
                depends_on: null
            # order: ["start", "duration", "percent_done", "resources", "depends_on", "color"]
            Env setup: [2022-09-06, 1, 0, "Martin", "brief"]
"""

    _, pert_def = ProjectReader(project_str).build_project()

    pert = PertDrawer(pert_def)
    pert.draw()
    assert pathlib.Path("test_project_3_pert.pdf").is_file()
    os.unlink("test_project_3_pert.gv")
    os.unlink("test_project_3_pert.pdf")


def test_patch_bug_0C501D():
    """
    :BUGFIX: 0C501D pertdrawer returns pert graphviz drawing string to stdout and gets reflected in the hidden output of the project definition

    - try suppressing the output in a contextlib function

    manual test: SUCCESS

    Status: CLOSED
    """
    pass


def test_patch_bug_39651C():
    """
    :BUGFIX: 39651C when a simple list definition is put in a project definition, the task is not always recognised.
    """
    project_str = """
# Font:
#     fill: black
#     font_family: Verdana
Vacations:
    - 2022-09-30
    - 2022-11-01
    - 2022-11-11
Resources:
    Martin:
        price:
            value: 68.75
            unit: workhour
        vacations:
            - 2022-09-15
    Samuel:
        price:
            value: 600
            unit:  workday
Projects:
    -   Name: Test1
        Tasks:
            kickoff:
                brief: [2022-09-05, 0.125, 0, null, null] # is strictly equivalent to common nested format
            goals:
                start: 2022-09-05
                duration: 0.25
                percent_done: 0
                resources: Martin
                depends_on: brief
            Env setup: [2022-09-06, 1, 0, "Martin", "goals"]
"""
    proj = yaml.full_load(project_str)
    assert is_nested(proj["Projects"][0]["Tasks"])


def test_patch_bug_406D86():
    """
    :BUGFIX: 406D86 nested simple list task definition raises error

    unsupported operand type(s) for *: 'int' and 'NoneType'

    """
    project_str = """
# Font:
#     fill: black
#     font_family: Verdana
Vacations:
    - 2022-09-30
    - 2022-11-01
    - 2022-11-11
Resources:
    Martin:
        price:
            value: 68.75
            unit: workhour
        vacations:
            - 2022-09-15
    Samuel:
        price:
            value: 600
            unit:  workday
Projects:
    -   Name: Test1
        Tasks:
            kickoff:
                brief: [2022-09-05, 0.125, 0, null, null] # is strictly equivalent to common nested format
            goals:
                start: 2022-09-05
                duration: 0.25
                percent_done: 0
                resources: Martin
                depends_on: brief
            Env setup: [2022-09-06, 1, 0, "Martin", "goals"]
"""
    project_def = ProjectReader(project_str)
    assert project_def


def test_path_bug_26BEBE():
    """
    :BUGFIX: 26BEBE nested uncertain task list definition raises

    TypeError: type NoneType doesn't define __round__ method

    issue in walk_project_tasks function => make sure the number of arguments in the list are 9. The item at index 6 is the color argument that must be a color html hex value or null.

    """
    project_str = """
# Font:
#     fill: black
#     font_family: Verdana
Vacations:
    - 2022-09-30
    - 2022-11-01
    - 2022-11-11
Resources:
    Martin:
        price:
            value: 68.75
            unit: workhour
        vacations:
            - 2022-09-15
    Samuel:
        price:
            value: 600
            unit:  workday
Projects:
    -   Name: Test1
        Tasks:
            kickoff:
                brief: [2022-09-05, null, 0, null, null, null, 1, 2, 3] # is strictly equivalent to common nested format
            goals:
                start: 2022-09-05
                duration: 0.25
                percent_done: 0
                resources: Martin
                depends_on: brief
            Env setup: [2022-09-06, null, 0, "Martin", "goals", null,  1, 2, 3]
"""
    project = yaml.full_load(project_str)
    assert walk_project_tasks(project["Projects"][0]["Tasks"])


def test_path_bug_778FE8():
    """
    :BUGFIX: 778FE8 nested tasks dependant on milestone don't resolve dependencies

    issue in walk_project_tasks function

    """
    project_str = """
# Font:
#     fill: black
#     font_family: Verdana
Vacations:
    - 2022-09-30
    - 2022-11-01
    - 2022-11-11
Resources:
    Martin:
        price:
            value: 68.75
            unit: workhour
        vacations:
            - 2022-09-15
    Samuel:
        price:
            value: 600
            unit:  workday
Projects:
    -   Name: Test1
        Tasks:
            documentation:
                    document items:
                        # [start, duration, percent_done , resources, depends_on]
                        item1 : [2022-10-03, 0.125, 0, "Martin", null]
                        # item2: [2022-09-06, 1, 0, "Martin", "item1"]
                        item2: [null, 1, 0, "Martin", "item1"]
                        item3 : [null, 1, 0, "Martin", "item2"]
                    debug model: [null, null, 0, "Martin", "document items", null, 2, 5, 10]
                    output model: [null, null, 0, "Martin", "debug model", null, 0.125, 0.125, 3]
"""
    project = yaml.full_load(project_str)
    res = walk_project_tasks(project["Projects"][0]["Tasks"])
    assert res[5][0] == "debug model"
    assert res[5][1]["depends_on"] == "document items"
    assert res

################################################################################
# CLI
################################################################################


@pytest.mark.skip("Not Implemented")
def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.main)
    assert result.exit_code == 0
    assert "md_planning.cli.main" in result.output
    help_result = runner.invoke(cli.main, ["--help"])
    assert help_result.exit_code == 0
    assert "--help  Show this message and exit." in help_result.output



"""md_planning main module."""
from datetime import date, datetime, timedelta
import pathlib
import re
import shutil
from typing import Any, Tuple, Union, Dict, List
from functools import singledispatch
from copy import deepcopy
import networkx as nx
from gantt import Resource, Task, Project, Milestone
import gantt
from criticalpath import Node
import yaml
from dateutil.parser import parse as parsedate, ParserError
import pertchart
import pint
import pandas as pd
from math import ceil

from contextlib import contextmanager, redirect_stderr, redirect_stdout
from os import devnull

# import debugpy
# debugpy.listen(5678)
# debugpy.wait_for_client()
# debugpy.breakpoint()

##### Work Units #####

# tailored pint unit registry (overridable) for project management
UREG = pint.UnitRegistry()
# unit = ""  #dimensionless unit to define things that come as units
UREG.define("unit = [dimensionless] = _ = u")
# batch = "" #dimensionless unit to define things that come in a group of units
UREG.define("@alias unit = batch = b")

UREG.define("workhour = 1 * hour = _ = wh = workhours")
assert "workhour" in UREG
assert "wh" in UREG

UREG.define(
    "workday = 8 * workhour = _ = wd = workdays"
)  # average number of working hours during the day
assert "workday" in UREG
assert "wd" in UREG

UREG.define(
    "workweek = 5 workday = _ = ww = workweeks"
)  # average number of working days during a week
assert "workweek" in UREG
assert "ww" in UREG

UREG.define(
    "workmonth = 4.33 workweek = _ = wm = workmonths"
)  # average number of working weeks in a month
assert "workmonth" in UREG
assert "wm" in UREG

UREG.define("workquarter = 3 * workmonth = _ = wq = workquarters")
assert "workquarter" in UREG
assert "wq" in UREG

# UREG.define("holiday = 6 * workweek = _ = holidays")
# assert "holiday" in UREG

# UREG.define("workyear = 52 * workweek; offset: -1 * holiday = _  = wy = workyears")
# assert "workyear" in UREG
# assert "wy" in UREG


@contextmanager
def suppress_stdout_stderr():
    """suppress_stdout_stderr is a context manager that redirects stdout and stderr to devnull."""
    with open(devnull, "w") as fnull:
        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
            yield (err, out)


################################################################################
# Patching Resource class
################################################################################
def _resource_get_registry(self) -> pint.UnitRegistry:
    return UREG


Resource.get_registry = _resource_get_registry


def _resource_get_unit(self):
    return self.price["unit"]


Resource.get_unit = _resource_get_unit


def _resource_get_price(self):
    return float(self.price["value"])


Resource.get_price = _resource_get_price


def _resource_convert(
    self, unit: str = "workday", number_of_days: Union[int, float, None] = None
) -> float:
    """
    convert finds the price of a resource as it is used on a daily basis.

    convert brings all resource, whichever unit, to a per day price for the creation of a budget for a project

    :param unit: the granularity of the analysis, defaults to "workday"
    :type unit: str, optional
    :param number_of_days: the total length of the associated task when the price of the resource is independant from time, defaults to None
    :type number_of_days: Union[int, float, None], optional
    :return: the price in the user's currency unit
    :rtype: float
    """
    Q_ = self.get_registry()

    if self.get_unit() == "unit":
        quantity = Q_(f"{self.get_price()} / {number_of_days} / workday")
    elif self.get_unit() == "batch":
        quantity = Q_(
            f"{self.get_price()} / {self.price['batch_size']} / {number_of_days} / workday"
        )
    else:
        quantity = Q_(f" {self.get_price()} / {self.get_unit()}")

    return quantity.to(f"1 / {unit}").m


Resource.convert = _resource_convert


################################################################################
# Patching the Task class
################################################################################


def _is_using(self, resource: str) -> bool:
    return resource in [r.name for r in self.resources]


Task.is_using = _is_using


def _is_active(self, date: Union[str, datetime.date]) -> bool:
    d = date
    if isinstance(d, str):
        try:
            d = parsedate(date).date()
        except ParserError:
            raise ValueError(f"Unrecognised Date: {date}")
    return self.start_date() <= d <= self.end_date()


Task.is_active = _is_active
################################################################################
# Utility functions
################################################################################


def _normalize_resources(task: dict) -> dict:
    res = deepcopy(task)
    if "resources" in res and isinstance(res["resources"], str):
        if "," in res["resources"]:  # type: ignore
            res["resources"] = list(map(lambda s: s.strip(), res["resources"].split(",")))
        else:
            res["resources"] = [res["resources"].strip()]

        regex = re.compile(r"^(\d+\.?\d*)\s+(\w+)$")
        for ind, resource in enumerate(res["resources"]):
            if regex.match(resource) is None:
                res["resources"][ind] = f"1 {resource}"

    if "resources" in res and res["resources"] is None:
        res["resources"] = []

    return res


def _normalize_duration(task: dict) -> dict:
    res = deepcopy(task)
    if res.get("duration") is None:  # then best, optimal and worst are defined
        res["duration"] = round(
            (res["best"] + 4 * res["optimal"] + res["worst"]) / 6, 3
        )
    else:  # no variability in the duration
        res["best"] = res["optimal"] = res["worst"] = res["duration"]
    return res


def _normalize_depends_on(task: dict) -> dict:
    # make sure parent tasks are in the form of List[str]
    res = deepcopy(task)
    if "depends_on" in res:
        if isinstance(res["depends_on"], str):
            if "," in res["depends_on"]:
                res["depends_on"] = list(
                    map(lambda s: s.strip(), res["depends_on"].split(","))
                )
            else:
                res["depends_on"] = [res["depends_on"].strip()]
        elif res["depends_on"] is None:
            res["depends_on"] = []
        else:
            pass
    return res


def _normalize_list_task(task: Union[list, dict]) -> dict:
    if isinstance(task, list):
        keys = {
            "start": None,
            "duration": None,
            "percent_done": 0,
            "resources": [],
            "depends_on": [],
            "color": None,
            "best": None,
            "optimal": None,
            "worst": None,
        }
        res = {"type": "task"}
        for ind, (k, default) in enumerate(keys.items()):
            try:
                res[k] = task[ind]
            except IndexError:
                res[k] = default
        return res
    elif isinstance(task, dict):
        return task
    else:
        raise ValueError(f"Unknown task definition {task}")


@singledispatch
def normalize_task(task: Any):
    """
    normalize_task changes a list defined task to a dict defined task.

    :param task: task defined as a list
    :type task: list
    :raises NotImplementedError: if the task type is not a recognised format
    """
    raise NotImplementedError


@normalize_task.register(dict)
def _(task: dict):
    return task


@normalize_task.register(list)
def _(task: list):
    keys = {
        "start": None,
        "duration": None,
        "percent_done": 0,
        "resources": [],
        "depends_on": [],
        "color": None,
        "best": None,
        "optimal": None,
        "worst": None,
        "is_milestone": False,
    }
    res = {}
    for ind, (k, default) in enumerate(keys.items()):
        try:
            res[k] = task[ind]
        except IndexError:
            res[k] = default
    return res


@singledispatch
def get_duration(task: Any):
    """
    get_duration makes the duration of a task primarily from the estimated task durations and if not available, the actual duration data.

    :param task: the task definition
    :type task: Union[dict, list]
    :raises NotImplementedType: types other than dict and list will raise error
    """
    raise NotImplementedError


@get_duration.register(dict)
def _(task: dict):
    if all(
        [
            task.get("best") is not None,
            task.get("optimal") is not None,
            task.get("worst") is not None,
        ]
    ):  # \BUGFIX 406D86
        return round((task["best"] + 4 * task["optimal"] + task["worst"]) / 6, 3)
    else:
        try:
            return round(task["duration"], 3)
        except KeyError:
            raise ValueError(
                "Task must define either ('best', 'optimal', 'worst') or 'duration'."
            )


@get_duration.register(list)
def _(task: list):
    # {
    #     "start": None,
    #     "duration": None,
    #     "percent_done": 0,
    #     "resources": [],
    #     "depends_on": [],
    #     "color": None,
    #     "best": None,
    #     "optimal": None,
    #     "worst": None,
    # }
    if len(task) == 9:
        return round((task[6] + 4 * task[7] + task[8]) / 6, 3)
    else:
        return round(task[1], 3)


@singledispatch
def _get_resources(res: Any):
    """
    _get_resources helper function that returns resources in a project compatible format.

    :param res: resource
    :type res: Union[str, list, None]
    :raises NotImplementedError: if the data is not in a recognised format
    :return: the list of resources used by the task
    :rtype: List[str]
    """
    if res is None:
        return []
    raise NotImplementedError


@_get_resources.register(str)
def _(res: str):
    if "," in res:
        return [r.strip() for r in res.split(",")]
    else:
        return [res.strip()]


@_get_resources.register(list)
def _(res: list):
    return res


@singledispatch
def get_resources(task: Any):
    """
    get_resources is a helper function that returns the resources used in a task in a project compatible format.

    Works on dict defined tasks and list defined tasks.

    :param task: the task description
    :type task: Union[dict, list]
    :raises NotImplementedError: if the task format is not recognised.
    :return: list of resource names
    :rtype: List[str]
    """
    raise NotImplementedError


@get_resources.register(dict)
def _(task: dict):
    return _get_resources(task["resources"])


@get_resources.register(list)
def _(task: list):
    return _get_resources(task[3])


@singledispatch
def _get_dependencies(deps: Any):
    if deps is None:
        return []
    raise NotImplementedError


@_get_dependencies.register(str)
def _(deps: str):
    if "," in deps:
        return list(map(lambda s: s.strip(), deps.split(",")))
    else:
        return [deps]


@_get_dependencies.register(list)
def _(deps: list):
    return deps


@singledispatch
def get_dependencies(task: Any):
    """
    get_dependencies get the dependencies from a task definition.

    Works on dict and list defined tasks

    :param task: the task definition
    :type task: Union[dict, list]
    :raises NotImplementedError: if the task definition is not recognised
    """
    raise NotImplementedError


@get_dependencies.register(dict)
def _(task: dict):
    return _get_dependencies(task["depends_on"])


@get_dependencies.register(list)
def _(task: list):
    return _get_dependencies(task[4])


def rename(
    dict_in: dict,
    key_from: str,
    key_to: str,
    flexible: bool = False,
    flex_val: Any = None,
) -> dict:
    """
    Forgiving function to rename keys in a dictionary.

    :dict_in:: the input dictionary
    :key_from:: the key to change
    :key_to:: the key to rename to
    :flexible:: raise error if key_from does not exist dict_in else assign None to key_to
    """
    if not flexible and key_from not in dict_in:
        raise KeyError(f"{key_from} not in input dictionary")
    data = deepcopy(dict_in)
    data[key_to] = data.get(key_from, flex_val)
    if key_from in data:
        del data[key_from]
    return data


@singledispatch
def get_task_type(item: Any) -> Union[str, None]:
    """
    get_task_type returns the name of the task type: milestone or task.

    :param item: item to be examined
    :type item: Any
    :raises NotImplementedError: if the item is different from a dict or list
    :return: the item tag type
    :rtype: str
    """
    return None


@get_task_type.register(dict)
def _(item: Dict):
    # flat list of tasks
    res = item.get("type")
    if res in ["task", "milestone"]:
        return res
    if "percent_done" in item:
        return "task"
    # nested list of milestones and tasks
    return "milestone"


@get_task_type.register(list)
def _(item: List):
    return "task"


def is_nested(tasks: dict) -> bool:
    """
    is_nested recognizes the project format to determine if the tasks definition is in the flat or the nested format.

    if any task is a milestone and contains another task then is_nested==True

    :param tasks: full project task definition
    :type tasks: dict
    :return: is it a nested definition T/F
    :rtype: bool

    :BUGFIX: E17FB6 and  39651C
    """
    for task in tasks.values():
        if get_task_type(task) == "task":
            continue
        else:  # is milestone
            for value in task.values():
                if get_task_type(value) in ["task", "milestone"]:
                    return True
    return False


def _walk(path, child):
    if isinstance(child, dict):
        if get_task_type(child) == "task":
            child["is_milestone"] = False
            yield [tuple(path), child]
        else:
            for key, value in child.items():
                if get_task_type(value) == "milestone":
                    yield [tuple(path + [key]), {"is_milestone": True}]
                yield from _walk(path + [key], value)
    else:
        yield [tuple(path), child]


def walk_project_tasks(tasks: dict):
    """
    walk_project returns normalized tasks in a nested or flat project configuration

    :param tasks: project tasks
    :type tasks: dict
    :return: list of task_name, task_definition
    :rtype: List[Tuple(str, dict)]
    """
    if is_nested(tasks):
        res = [(pth, normalize_task(tsk)) for (pth, tsk) in list(_walk([], tasks))]

        milestones = list(
            sorted(
                filter(
                    lambda i: not isinstance(i[1], list) and i[1]["is_milestone"], res
                ),
                key=lambda m: len(m[0]),
                reverse=True,
            )
        )
        # milestone_ids = [i[0][-1] for i in milestones]

        _tasks = list(
            sorted(
                filter(
                    lambda i: isinstance(i[1], list) or not i[1]["is_milestone"], res
                ),
                key=lambda m: len(m[0]),
                reverse=True,
            )
        )

        # replace all milestones depends_on that depend on a milestone or a group of tasks
        for milestone in milestones:
            milestone_path_len = len(milestone[0])
            # project = Node(milestone[0][-1])
            subtasks = list(
                filter(
                    lambda task: len(task[0]) == milestone_path_len + 1
                    and task[0][-2] == milestone[0][-1],
                    res,
                )
            )

            # for subtask in subtasks:  # 1 level below milestone and subtask of milestone
            #     try:
            #         if subtask[1]["is_milestone"]:
            #             dep_names = get_dependencies(subtask[1])
            #             for dep in dep_names:
            #                 _t = list(filter(lambda t: t[0][-1]==dep, _tasks))[0]
            #                 project.add(
            #                     Node(dep, duration=get_duration(subtask[1]), lag=0)
            #                 )
            #         else:
            #             project.add(
            #                 Node(
            #                     subtask[0][-1],
            #                     duration=get_duration(subtask[1]),
            #                     lag=0,
            #                 )
            #             )
            #     except Exception as err:
            #         raise ValueError(f"Definition error at {subtask[0][-1]}") from err

            # add dependencies
            # for subtask in subtasks:
            #     if (deps:=get_dependencies(subtask[1])):
            #         updated_deps = []

            #         for dep in deps: # not dependency but parent really...
            #             # update milestone dependency to last node in critical path
            #             if dep in milestone_ids:
            #                 updated_deps.append(get_dependencies(list(filter(lambda m: m[0][-1] == dep, milestones))[0][1])[0]) # there is only one last task on critical path
            #             else:
            #                 updated_deps.append(dep)

            #         # filter updated_deps to tasks strictly within descendants of the current milestone
            #         # subtasks_names = [s[0][-1] for s in subtasks]
            #         # updated_deps = [d for d in updated_deps if d in subtasks_names]

            #         # build project links
            #         for _d in updated_deps:
            #             try:
            #                 project.link(_d, subtask[0][-1])
            #             except Exception as err:
            #                 # raise ValueError(f"Cannot link {subtask[0][-1]} and {_d}") from err
            #                 node = Node(
            #                     _d,
            #                     duration=0,
            #                     lag=0,
            #                 )
            #                 project.add(node)
            #                 project.link(_d, subtask[0][-1])

            # project.update_all()
            # critical_path = list(map(lambda i: i.name, project.get_critical_path()))
            # milestone[1]["depends_on"] = [critical_path[0], critical_path[-1]]
            milestone[1]["depends_on"] = [subtask[0][-1] for subtask in subtasks]
            # milestone[1]["duration"] = project.duration
            milestone[1]["duration"] = 0
            # milestone[1]["first_node"] = critical_path[0]
            # milestone[1]["last_node"] = critical_path[-1]

        # # replace all tasks that depend on a milestone to its last node
        # for task in _tasks:
        #     try:
        #         dependencies = get_dependencies(task[1])
        #         # make a temporary result list
        #         dependency_results = []
        #         # for each item in dependencies
        #         for dependency in dependencies:
        #             #if item is a milestone and milestone in task['depends_on']
        #             if dependency in milestone_ids:
        #                 # get the milestone dependency
        #                 _milestone = list(filter(lambda m: m[0][-1]==dependency, milestones))[0]
        #                 # extend temp result list with this milestone's dependencies
        #                 dependency_results.extend(get_dependencies(_milestone[1]))
        #             # else:
        #             else:
        #                 dependency_results.append(dependency)
        #                 #extend the temp result list with the current dependency
        #         task[1]["depends_on"] = dependency_results
        #         # task[1]["depends_on"] = [ get_dependencies(list(filter(lambda _milst: _milst[0][-1]==m,milestones))[0][1]) if m==d else d for d in dependencies for m in milestone_ids]
        #     except Exception as err:
        #         raise ValueError(f"Dependency error at {task[0][-1]}")

        return [(task[0][-1], task[1]) for task in res]
    else:
        return list(tasks.items())


################################################################################
# Project reader class
################################################################################
class ProjectReader:
    """
    Utility for reading raw project data.

    Reads yaml data and returns a representation of the projects
    suitable for a PERT and gantt analysis.
    """

    def __init__(self, projstr: str, critical_color: Union[str, None] = None):
        """
        Read and setup project data.

        :projstr:: string representation of the project in yaml syntax
        :critical_color:: string/name/hex representation of the color for the tasks on the critical path
        """

        self.projstr = projstr
        self.data = yaml.safe_load(projstr)
        self.font = self.data.get("Font", {})
        self.vacations = self.data.get("Vacations", [])
        self.resources = self.data.get("Resources", {})
        self.projects = {}
        self.tasks = {}
        self.pertcharts = {}

        for project in self.data.get("Projects", []):  # type: ignore
            proj_name = project.get("Name")
            if not proj_name:
                raise ValueError("Missing project name")
            self.projects[proj_name] = None
            tasks: dict = project.get("Tasks")

            for name, task in walk_project_tasks(tasks):

                ##### SETUP #####

                payload = {
                    "project": proj_name,
                    "task": {
                        "name": name,
                        "start": None,
                        "stop": None,
                        "duration": None,
                        "depends_of": None,
                        "resources": None,
                        "percent_done": 0,
                        "color": None,
                        "fullname": None,
                        "display": True,
                        "state": "",
                    },
                    "milestone": {
                        "name": name,
                        "start": None,
                        "depends_of": None,
                        "color": None,
                        "fullname": None,
                        "display": True,
                    },
                    "pertchart": {},
                    "cpm": {
                        "name": name,
                        "duration": None,
                        "lag": 0,
                        "depends_of": None,
                    },
                    "is_milestone": False,
                }

                ##### TRANSFORM #####

                task = _normalize_list_task(task)

                if (
                    task.get("is_milestone", False)
                    or task.get("type", "task") == "task"
                ):
                    try:
                        task = _normalize_duration(task)
                    except (TypeError, KeyError) as err:
                        raise ValueError(
                            f"Definition error for duration or estimates in task named {name}"
                        ) from err
                else:
                    task["duration"] = 0

                task = _normalize_depends_on(task)

                task = _normalize_resources(task)

                ##### UPDATE #####

                ##### CPM #####
                payload["cpm"]["name"] = name
                payload["cpm"]["duration"] = task["duration"]
                payload["cpm"]["lag"] = 0
                payload["cpm"]["depends_of"] = task.get("depends_on", None)

                ##### PERTCHART #####
                payload["pertchart"]["Tid"] = name
                payload["pertchart"]["start"] = 0
                payload["pertchart"]["duration"] = task["duration"]
                payload["pertchart"]["end"] = 0
                payload["pertchart"]["responsible"] = ""
                payload["pertchart"]["pred"] = task.get("depends_on", ["START"])
                if not payload["pertchart"]["pred"]:
                    payload["pertchart"]["pred"] = ["START"]

                ##### TASK #####
                for k in payload["task"]:
                    if k != "depends_on":
                        payload["task"][k] = task.get(k, payload["task"][k])
                payload["task"]["depends_of"] = task.get("depends_on", None)

                ##### MILESTONE #####
                if (
                    task.get("is_milestone", False)
                    or task.get("type", "task") == "milestone"
                ):
                    payload["is_milestone"] = True

                    for k in payload["milestone"]:
                        if k != "depends_on":
                            payload["milestone"][k] = task.get(
                                k, payload["milestone"][k]
                            )
                    payload["milestone"]["depends_of"] = task.get("depends_on", [])

                self.tasks[name] = payload

            ##### State: Ready for post-processing #####

            # CPM for PertChart
            critical_path = self._get_critical_path(
                {
                    k: t["cpm"]
                    for (k, t) in self.tasks.items()
                    if t["project"] == proj_name
                }
            )

            for task_name in critical_path:
                self.tasks[task_name]["pertchart"]["responsible"] = "CRITICAL"
                if self.tasks[task_name]["task"]["color"] is None:
                    self.tasks[task_name]["task"]["color"] = critical_color

    def _get_critical_path(self, cpm_data: dict) -> tuple:
        """
        _get_critical_path returns tuple of task (node) names on the critical path.

        :param cpm_data: cpm data as provided by the initialization
        :type cpm_data: dict
        :return: tuple of task names
        :rtype: tuple
        """
        proj = Node("project")
        tmp = {}
        for k, values in cpm_data.items():  # add nodes
            # cpm_data[k]["node"] = Node(
            tmp.setdefault(k, {})["node"] = Node(
                values["name"], duration=values["duration"], lag=values["lag"]
            )
            proj.add(tmp[k]["node"])

        for k, values in cpm_data.items():  # add links
            for link in values["depends_of"]:
                # if k and link and tmp:
                proj.link(tmp[link]["node"], tmp[k]["node"])

        proj.update_all()

        return tuple(map(str, proj.get_critical_path()))  # type: ignore

    def build_project(self) -> tuple:
        """
        build_project returns a tuple with  gantt project data and pertchart data.

        :return: pos 0 gantt data, pos 1 pert chart data
        :rtype: tuple
        """
        project_data = {
            "font": self.font,
            "vacations": self.vacations,
            "resources": self.resources,
            "projects": self.projects,
            "tasks": self.tasks,
        }

        pertchart_data = {
            k: {
                t["pertchart"]["Tid"]: t["pertchart"]
                for t in self.tasks.values()
                if t["project"] == k
            }
            for k in self.projects
        }

        return (project_data, pertchart_data)


################################################################################
# Pert drawer class
################################################################################
class PertDrawer:
    """
    PertParser takes a python dictionary in and extracts the pert data.

    from it in order to assess project CPM.

    It also cleans project data of PERT estimates in order to provide
    the proper durations for a gantt representation.
    """

    def __init__(self, tasks: dict):
        """
        Class initialization.

        :param tasks: pert compatible tasks definitions
        :type tasks: dict
        """
        self.tasks: dict = tasks

    def draw(
        self, project: Union[str, None] = None, out: Union[str, None] = None
    ) -> None:
        """
        draw Draws the network of tasks in self.tasks.

        :param project: name of the project otherwise all projects tasks will be analysed together, defaults to None
        :type project: Union[str, None], optional
        :param out: prefix name of the files to output, will default to the name of the project to output "project_pert.pdf" and "project_pert.gv" (graphviz file)
        :type out: Union[str, None], optional
        """
        if project is None:
            _projects = [k for k in self.tasks]
        else:
            # XXX: this is not of the same type as the definition above...
            _projects = [project]

        pert = pertchart.PertChart()

        for proj in _projects:
            calculated = pert.calculate_values(self.tasks[proj])

            # see bugfix 0C501D
            with suppress_stdout_stderr():
                pert.create_pert_chart(calculated)

            # see bugfix A14E36
            if pathlib.Path("PERT.gv.pdf").is_file():
                shutil.move(
                    "PERT.gv.pdf", f"{out}_pert.pdf" if out else f"{proj}_pert.pdf"
                )
            if pathlib.Path("PERT.gv").is_file():
                shutil.move("PERT.gv", f"{out}_pert.gv" if out else f"{proj}_pert.gv")

    # XXX: (refactor) not evidently useful usecase
    def get_critical_path(self, project: str):
        """
        get_critical_path returns the critical path of tasks in a specific project.

        :param project: name of the project
        :type project: str
        :raises NotImplementedError: _description_
        """
        raise NotImplementedError


################################################################################
# Gantt drawer class
################################################################################
class GanttDrawer:
    """
    Gantt project representation from a yaml file definition.

    see tests for project definition structure and class usage

    """

    def __init__(self, data):
        """
        Class initialization.

        :param data: definition of the project as output by ProjectReader class
        :type data: dict
        """
        self.registry = UREG
        self.data = data
        self.projects: Dict[str, Project] = {}
        self.resources: Dict[str, Resource] = {}
        self.vacations = []
        self.tasks: Dict[str, Task] = {}
        # self._register_font_attribute() #svgwrite issue
        self._register_projects()
        self._register_resources()
        self._register_vacations()
        self._register_tasks()

    def _register_font_attribute(self):
        """_register_font_attributes changes the default presentation settings."""
        # BUG: for whatever reason, the gantt API does not seem to work there...
        gantt.define_font_attributes(self.data["font"])

    def _register_projects(self) -> None:
        """_register_projects adds all projects and monkey patches class api."""
        self.projects = {k: Project(k) for k in self.data["projects"]}

    def _register_resources(self) -> None:
        """_register_resources adds and monkey patches class api."""
        self.resources = {}

        for resource in self.data["resources"]:
            try:
                res = Resource(resource)
                res.price = self.data["resources"][resource].get(
                    "price", {"value": 0, "unit": "unit"}
                )

                if res.price["unit"] == "batch":
                    if "batch_size" not in res.price:
                        raise ValueError(
                            f"Error in Resource {res.name}: batch resource definition missing key 'batch_size'"
                        )
                if (
                    vacations := self.data["resources"][resource].get("vacations")
                ) is not None:
                    res.add_vacations(*vacations)

                if res.price["unit"] in ["unit", "batch"] and res.vacations:
                    raise ValueError(
                        f"Resource:{res.name} cannot have unit/batch unit type and holidays at the same time"
                    )
            except Exception as err:
                raise ValueError(f"Definition error in {resource}") from err

            self.resources[resource] = res

    def _register_vacations(self) -> None:
        """_register_vacations adds vacations at a multiproject level."""
        if (vacations := self.data.get("vacations")) is not None:
            for _v in vacations:
                gantt.add_vacations(_v)

    def _register_tasks(self) -> None:
        """
        _register_tasks creates the network of tasks to register for the projects.

        get_tasks starts with the root task then iterates over all the left
        over tasks in search of the dependant tasks. Task names must be
        unique so it is safe to refer to them verbatim.
        """
        tasks = [
            {
                "project": t["project"],
                "task_name": t["is_milestone"]
                and t["milestone"]["name"]
                or t["task"]["name"],
                "task": t["milestone"] if t["is_milestone"] else t["task"],
                "is_milestone": t["is_milestone"],
                "depends_of": t["is_milestone"]
                and t["milestone"]["depends_of"]
                or t["task"]["depends_of"],
            }
            for t in self.data["tasks"].values()
        ]

        for proj_name, project in self.projects.items():
            task_registry = {}
            filtered_tasks = list(filter(lambda t: t["project"] == proj_name, tasks))
            linked_tasks = [
                (pred, task["task_name"])
                for task in filtered_tasks
                for pred in task["depends_of"]
            ]

            DG = nx.DiGraph(linked_tasks)

            for _t_name in nx.topological_sort(DG):
                to_register = list(
                    filter(lambda t: t["task_name"] == _t_name, filtered_tasks)
                )[0]
                dependents = [task_registry[name] for name in to_register["depends_of"]]
                resources = [
                    self.resources[res_check]
                    for res in to_register["task"].get("resources", [])
                    for res_check in self.resources
                    if res_check in res
                ]
                if to_register["is_milestone"]:
                    task_registry[_t_name] = Milestone(
                        **{**to_register["task"], "depends_of": dependents}
                    )
                else:
                    task_registry[_t_name] = Task(
                        **{
                            **to_register["task"],
                            "depends_of": dependents,
                            "resources": resources,
                        }
                    )
                project.add_task(task_registry[_t_name])

            self.tasks.update(**task_registry)

    def draw_tasks(
        self,
        project: str,
        filename: Union[str, None] = None,
        today: Union[datetime.date, str, None] = None,  # type: ignore
        start: Union[datetime.date, str, None] = None,  # type: ignore
        end: Union[datetime.date, str, None] = None,  # type: ignore
        scale: Union[str, None] = None,  # type: ignore
    ) -> None:
        """
        draw_tasks creates a svg visualization of the task execution in the project.

        :param project: project name
        :type project: str
        :param filename: file name override. If None, use the project name, defaults to None
        :type filename: Union[str, None], optional
        :param today: today's date for task burn down evaluation, defaults to None
        :type today: Union[datetime.date, str, None], optional
        """
        if filename is None:
            filename = f"{project}.svg"
        if today == "today":
            today = datetime.today().date()
        if isinstance(today, str):
            today = parsedate(today).date()
        if isinstance(start, str):
            start = parsedate(start).date()
        if isinstance(end, str):
            end = parsedate(end).date()

        scales = {
            "d": gantt.DRAW_WITH_DAILY_SCALE,
            "daily": gantt.DRAW_WITH_DAILY_SCALE,
            "w": gantt.DRAW_WITH_WEEKLY_SCALE,
            "weekly": gantt.DRAW_WITH_WEEKLY_SCALE,
            "m": gantt.DRAW_WITH_MONTHLY_SCALE,
            "monthly": gantt.DRAW_WITH_MONTHLY_SCALE,
            "q": gantt.DRAW_WITH_QUATERLY_SCALE,
            "quarterly": gantt.DRAW_WITH_QUATERLY_SCALE,
        }
        if scale is None:
            scale = "d"
        scale: str = scales[scale]

        self.projects[project].make_svg_for_tasks(
            filename=filename, today=today, start=start, end=end, scale=scale
        )

    def draw_resources(
        self,
        project: str,
        filename: Union[str, None] = None,
        today: Union[datetime.date, str, None] = None,  # type: ignore
        start: Union[datetime.date, str, None] = None,  # type: ignore
        end: Union[datetime.date, str, None] = None,  # type: ignore
        resources: Union[List[str], None] = None,  # type: ignore
        one_line_for_tasks: bool = False,
        scale: Union[str, None] = None,  # type: ignore
    ) -> None:
        """
        draw_resources creates a svg visualization of the resource use in the project.

        :param project: name of the project
        :type project: str
        :param filename: file name override. If None use the project name, defaults to None
        :type filename: Union[str, None], optional
        :param today: today's date for resource use WIP evaluation, defaults to None
        :type today: Union[datetime.date, str, None], optional
        :param scale: python_gantt scale enumeration see doc, defaults to None
        :type scale: Union[str, None], optional
        """
        if filename is None:
            filename = f"{project}_resources.svg"
        if today == "today":
            today = datetime.today().date()
        if isinstance(today, str):
            today = parsedate(today).date()
        if isinstance(start, str):
            start = parsedate(start).date()
        if isinstance(end, str):
            end = parsedate(end).date()
        if resources:
            resources: Resource = [self.resources[res] for res in resources]  # type: ignore
        scales = {
            "d": gantt.DRAW_WITH_DAILY_SCALE,
            "daily": gantt.DRAW_WITH_DAILY_SCALE,
            "w": gantt.DRAW_WITH_WEEKLY_SCALE,
            "weekly": gantt.DRAW_WITH_WEEKLY_SCALE,
            "m": gantt.DRAW_WITH_MONTHLY_SCALE,
            "monthly": gantt.DRAW_WITH_MONTHLY_SCALE,
            "q": gantt.DRAW_WITH_QUATERLY_SCALE,
            "quarterly": gantt.DRAW_WITH_QUATERLY_SCALE,
        }
        if scale is None:
            scale = "d"
        scale: str = scales[scale]

        self.projects[project].make_svg_for_resources(
            filename=filename,
            today=today,
            start=start,
            end=end,
            resources=resources,
            one_line_for_tasks=one_line_for_tasks,
            scale=scale,
        )

    def define_unit(self, unit: str) -> None:
        """
        define_unit registers a new pint unit in the instance unit registry.

        :param unit: unit definition (see pint documentation for syntax)
        :type unit: str
        """
        self.registry.define(unit)

    def define_alias(self, alias: str) -> None:
        """
        define_alias makes a new alias for an existing unit.

        useful when using a unit in a multilingual and pluralized context

        :param alias: alias definition, can start with "@alias" (full definition) or not (abbreviated syntax) (see pint documentation for more details)
        :type alias: str
        """
        if alias.strip().startswith("@alias"):
            self.registry.define(alias)
        else:
            self.registry.define(f"@alias {alias}")

    def get_unit(self, resource: str) -> str:
        """
        get_unit helper function for a project object to get a resource's units

        :param project: name of the project
        :type project: str
        :param resource: name of the resource
        :type resource: str
        """
        try:
            return self.resources[resource].get_unit()
        except KeyError:
            raise ValueError(f"Unknown resource {resource}")

    def get_price(self, resource: str) -> float:
        """
        get_price returns the price of the specified resource

        :param resource: name of the resource
        :type resource: str
        :return: the price of the resource
        :rtype: float
        """
        try:
            return self.resources[resource].get_price()
        except KeyError:
            raise ValueError(f"Unknown resource {resource}")

    def get_usage(self, task: str, resource: str) -> float:
        """
        get_usage returns the amount of a specified resource that a task uses.

        :param task: task name
        :type task: str
        :param resource: resource name
        :type resource: str
        :raises ValueError: if the task name is not defined
        :raises ValueError: if the resource name is not defined
        :return: the amount used of specified resource by a task
        :rtype: float
        """
        try:
            resources = self.data["tasks"][task]["task"]["resources"]
        except KeyError:
            raise ValueError(f"Unknown task {task}")

        if resource not in self.resources:
            raise ValueError(f"Unknown resource {resource}")

        for res in resources:
            if res.endswith(resource):
                return float(res.split()[0])

        return 0.0

    def get_bounds(self) -> Tuple[date, date]:
        """
        get_bounds returns the start date and the end date of the whole file definition, all projects taken into consideration.

        :return: start_date, end_date
        :rtype: Tuple[date, date]
        """
        start = None
        stop = None
        for proj in self.projects.values():
            if start is None or start > proj.start_date():
                start = proj.start_date()
            if stop is None or stop < proj.end_date():
                stop = proj.end_date()
        return start, stop

    def is_available(self, resource: str, date: Union[str, datetime.date]) -> bool:
        """
        is_available returns True if a specific resource is available to use on a specified date.

        :param resource: resource name
        :type resource: str
        :param date: date at which availability is assessed
        :type date: Union[str, datetime.date]
        :raises ValueError: _description_
        :return: True/False
        :rtype: bool
        """
        start, stop = self.get_bounds()

        try:
            res = self.resources[resource]
        except KeyError:
            raise ValueError(f"Resource named {resource} not defined")

        d = date
        if isinstance(d, str):
            try:
                d = parsedate(date).date()
            except ParserError:
                raise ValueError(f"Unrecognised Date: {date}")

        if start <= d <= stop:
            return res.is_available(d)

        return False

    def is_using(self, task: str, resource: str) -> bool:
        """
        is_using is a utility function at the project level to determine if the given task is using a specific resource.

        :param task: task name
        :type task: str
        :param resource: resource name
        :type resource: str
        :return: is task ABC using resource XYZ, True/False
        :rtype: bool
        """
        return self.tasks[task].is_using(resource)

    def convert(
        self, resource: str, number_of_days: Union[int, float, None] = None
    ) -> float:
        """
        convert is a utility function at the project level that gets the daily price of a resource with units.

        :param resource: resource name
        :type resource: str
        :param number_of_days: length of the associated task for resources that are time independant, defaults to None
        :type number_of_days: Union[int, float, None], optional
        :return: a price in the user's currency unit
        :rtype: float
        """
        return self.resources[resource].convert(number_of_days=number_of_days)

    def is_active(self, task: str, date: Union[str, datetime.date]) -> bool:
        """
        is_active is a utility function at the project level that gets the activity status of a task in a project at a given date.

        :param task: task name
        :type task: str
        :param date: the date when the activity is considered
        :type date: Union[str, datetime.date]
        :raises ValueError: if the date is not recognised by dateutil.parse
        :return: the task is active: True/False
        :rtype: bool
        """
        d = date
        if isinstance(d, str):
            try:
                d = parsedate(date).date()
            except ParserError:
                raise ValueError(f"Unrecognised Date: {date}")

        return self.tasks[task].is_active(d)

    def duration(self, task: str) -> float:
        """
        duration is a utility function at the project level to retrieve a task duration.

        :param task: task name
        :type task: str
        :return: task duration
        :rtype: float
        """
        return self.tasks[task].duration

    def get_cost(
        self, task: str, resource: str, date: Union[str, datetime.date]
    ) -> float:
        """
        get_cost returns the actual cost of a specific resource in a given task, on a given day.

        :param task: task name
        :type task: str
        :param resource: resource name
        :type resource: str
        :param date: date of consideration
        :type date: Union[str, datetime.date]
        :raises ValueError: if the date format is not recognised by dateutil.parse
        :return: the cost in the user's currency
        :rtype: float
        """
        d = date
        if isinstance(d, str):
            try:
                d = parsedate(date).date()
            except ParserError:
                raise ValueError(f"Unrecognised Date: {date}")

        task_length = (mytask := self.tasks[task]).duration or float(
            mytask.end_date() - mytask.start_date()
        )  # type: ignore

        task_days, task_decimals = (task_length // 1, task_length % 1)

        duration_left = (mytask.start_date() + timedelta(days=task_days)) - d

        if duration_left.days > task_days:  # task not active
            return 0

        time_scaling = 1 if duration_left.days != 0 else task_decimals

        return (
            self.convert(resource, number_of_days=ceil(task_length))
            * time_scaling
            * self.get_usage(task, resource)
            * self.is_available(resource, d)
            * self.is_using(task, resource)
            * self.is_active(task, d)
        )

    def budget(self) -> pd.DataFrame:
        """
        budget records all the planned expenses

        budget records for each day, each task and each resource, the use and associated cost in the project planning.

        :return: pandas.DataFrame with columns ["project", "task", "resource", "category", "subcategory", "date", "amount"]
        :rtype: pd.DataFrame
        """
        columns = [
            "project",
            "task",
            "resource",
            "category",
            "subcategory",
            "date",
            "amount",
        ]
        result = []
        start, stop = self.get_bounds()
        days = (stop - start).days + 1  # the end date is included in the bounds
        for proj_name, project in self.projects.items():
            for task in project.tasks:
                if hasattr(task, "resources"):
                    for resource in task.resources:
                        for ind in range(days):
                            now = start + timedelta(days=ind)
                            if (
                                amount := self.get_cost(task.name, resource.name, now)
                            ) != 0:
                                result.append(
                                    [
                                        proj_name,
                                        task.name,
                                        resource.name,
                                        "",
                                        "",
                                        now,
                                        amount,
                                    ]
                                )
        return pd.DataFrame(result, columns=columns)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "Click>=7.0",
    "pyyaml==6.0",
    "python-gantt==0.6.0",
    "python-gantt-csv==0.4.0",
    "pertchart==0.5.1",
    "criticalpath==0.1.5",
    "networkx==2.8.4",
    "pint==0.19.2",
    "pandas==1.4.3",
    "tabulate==0.8.10",
    "matplotlib==3.6.0",
]

test_requirements = [
    "pytest>=3",
]

setup(
    author="Martin Teller",
    author_email="martintellerpro@gmail.com",
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description="package to define, visualize with gantt and budget "
    "projects using yaml syntax. ",
    entry_points={
        "console_scripts": [
            "md_planning=md_planning.cli:main",
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="md_planning project gantt PERT CPM budgeting budget visualization",
    name="md_planning",
    packages=find_packages(include=["md_planning", "md_planning.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/mtllr/md_planning",
    version='0.5.0',
    zip_safe=False,
    project_urls={
        "Homepage": "https://github.com/mtllr/md_planning",
        "Documentation": "https://md-planning.readthedocs.io",
        "Buy me a coffee": "https://www.buymeacoffee.com/mtllr",
    },
)
